#! /usr/bin/python3
#
#################################################################
## Copyright (c) 2015 Norbert S. <junky-zs@gmx.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#################################################################
# Ver:0.1    / Datum 19.06.2017 first release
# Ver:0.2    / Datum  5.09.2018 Cxyz - Regler added
#################################################################

import sys, time
import threading
sys.path.append('lib')

import mqtt_client_if
import paho.mqtt.client as paho

import ht_utils, logging


__author__  = "Norbert S <junky-zs@gmx.de>"
__status__  = "draft"
__version__ = "0.4"
__date__    = "5.08.2018"

threadlock=threading.Lock()
#global used values
g_TSolarSpeicherMaxValue=82.0
g_TWatermarkLow =g_TSolarSpeicherMaxValue-12.0
g_TWatermarkMed =g_TSolarSpeicherMaxValue- 8.0
g_TWatermarkHigh=g_TSolarSpeicherMaxValue- 5.0
g_THysterese   =  5.0
g_TSollDefault = 21.5

# setup controller-type "Fxýz" for FWxyz of FRxyz controller
# setup controller-type "Cxýz" for CWxyz of CRxyz controller

g_ControllerType = "Cxyz"


class cdata():
    def __init__(self):
        #heater-values
        self.__i_tvorlauf_soll =0
        self.__i_tvorlauf_ist  =0.0
        self.__i_leistung      =0
        self.__b_brennerflamme= 0
        self.__f_tAussen       =0.0
        #heater-circuit-values
        self.__i_betriebsstatus   =0
        self.__f_Soll_HK       =0.0
        self.__f_Steuer_FB     =0.0
        self.__i_tempera_niveau=0
        #solar-values
        self.__f_kollektor     =0.0
        self.__f_speicherunten =0.0
        self.__b_pumpe         =0
        self.__i_speicher_voll =0
        self.__i_kollektor_aus =0
        self.__myLock = threading.Lock()

        #heater-methodes
    def i_tvorlauf_soll(self, i_tvorlauf_soll=-0):
        self.__myLock.acquire()
        if i_tvorlauf_soll != -0:
            self.__i_tvorlauf_soll = i_tvorlauf_soll
        self.__myLock.release()
        return self.__i_tvorlauf_soll

    def f_tvorlauf_ist(self, f_tvorlauf_ist=-0.0):
        self.__myLock.acquire()
        if f_tvorlauf_ist != -0.0:
            self.__i_tvorlauf_ist = f_tvorlauf_ist
        self.__myLock.release()
        return float(self.__i_tvorlauf_ist)

    def b_brennerflamme(self, b_brennerflamme = -1):
        self.__myLock.acquire()
        if b_brennerflamme != -1:
            self.__b_brennerflamme = b_brennerflamme
        self.__myLock.release()
        return self.__b_brennerflamme

    def f_tAussen(self, f_tAussen=-0.0):
        self.__myLock.acquire()
        if f_tAussen != -0.0:
            self.__f_tAussen = f_tAussen
        self.__myLock.release()
        return float(self.__f_tAussen)

        #heater circuit -methodes
    def i_betriebsstatus(self, i_betriebsart=-0):
        self.__myLock.acquire()
        if i_betriebsart != -0:
            self.__i_betriebsstatus = i_betriebsart
        self.__myLock.release()
        return self.__i_betriebsstatus

    def f_Soll_HK(self, f_Soll_HK=-0.0):
        self.__myLock.acquire()
        if f_Soll_HK != -0.0:
            self.__f_Soll_HK = f_Soll_HK
        self.__myLock.release()
        return float(self.__f_Soll_HK)

    def f_Steuer(self, f_Steuer=-0.0):
        self.__myLock.acquire()
        if f_Steuer != -0.0:
            self.__f_Steuer_FB = f_Steuer
        self.__myLock.release()
        return float(self.__f_Steuer_FB)

        #solar-methodes
    def f_kollektor(self, f_kollektor=-0.0):
        self.__myLock.acquire()
        if f_kollektor != -0.0:
            self.__f_kollektor = f_kollektor
        self.__myLock.release()
        return float(self.__f_kollektor)

    def f_speicherunten(self, f_speicherunten=-0.0):
        self.__myLock.acquire()
        if f_speicherunten != -0.0:
            self.__f_speicherunten = f_speicherunten
        self.__myLock.release()
        return float(self.__f_speicherunten)

    def b_pumpe(self, b_pumpe = -1):
        self.__myLock.acquire()
        if b_pumpe != -1:
            self.__b_pumpe = b_pumpe
        self.__myLock.release()
        return self.__b_pumpe

    def i_speicher_voll(self, i_speicher_voll = -1):
        self.__myLock.acquire()
        if i_speicher_voll != -1:
            self.__i_speicher_voll = i_speicher_voll
        self.__myLock.release()
        return self.__i_speicher_voll

    def i_kollektor_aus(self, i_kollektor_aus = -1):
        self.__myLock.acquire()
        if i_kollektor_aus != -1:
            self.__i_kollektor_aus = i_kollektor_aus
        self.__myLock.release()
        return self.__i_kollektor_aus

    def i_temperatur_niveau(self, i_tniveau = -1):
        self.__myLock.acquire()
        if i_tniveau != -1:
            self.__i_tempera_niveau = i_tniveau
        self.__myLock.release()
        return int(self.__i_tempera_niveau)

class my_cmqtt_client(mqtt_client_if.cmqtt_baseclass):
    """
    """

    def __init__(self, cfg_filename):
        mqtt_client_if.cmqtt_baseclass.__init__(self, cfg_filename)
        self.data = cdata()

    def __del__(self):
        pass

    def subscribe_settopics(self, subscribe_topic="hometop/ht/#"):
        """subscribes to the topic-parameter, default:'set/<topic_root_name>/#' is used."""
        #subscribe that topic
        (result, mid) = self._client_handle().subscribe(subscribe_topic, qos=1)
        if result != paho.MQTT_ERR_SUCCESS:
            #error occured un subscribing, log it
            errorstr = """my_cmqtt_client.subscribe_settopics() subscrib-error:{0}""".format(result)
#            print(errorstr)
            self.cfg_logging().critical(errorstr)

    def publish_data(self, topic, value):
        """sending data to known and connected broker."""
        try:
            (rc, mid) = self._client_handle().publish(topic,str(value), qos=self.cfg_QoS(), retain=self.cfg_RetainFlag())
            if rc != 0:
                errorstr = "cmqtt_client.publish_data() error occured; topic:{0}; mid:{1}".format(topic, mid)
#                print(errorstr)
                self.cfg_logging().warning(errorstr)
            else:
                debugstr = "publish topic {0}: {1}".format(topic, value)
                self.cfg_logging().info(debugstr)

                self._client_handle().will_set(topic, payload=value, qos=self.cfg_QoS(), retain=self.cfg_LastWillRetain())
        except:
            errorstr = "cmqtt_client.publish_data() error occured on topic:{0}".format(topic)
#            print(errorstr)
            self.cfg_logging().critical(errorstr)
            raise

    def mqtt_client_handle(self):
        return self

    def run(self):
        """ """
        self.cfg_read()
        init_OK = False

        try:
            # mqtt client initialisation
            init_OK = self.mqtt_init()
        except:
            errorstr = "my_cmqtt_client.mqtt_init() error"
            self.cfg_logging().critical(errorstr)
            raise SystemExit

        if init_OK == True:
            self._solarcontrol = csolarcontrol(self.mqtt_client_handle(), self.cfg_logging(), self.data)
            self.subscribe_settopics('hometop/ht/#')

        while self.loop() and init_OK:
            self._solarcontrol.run_solarcontrol()

        self.cfg_logging().critical("my_cmqtt_client.solarcoontrol unexpected terminated")

    def processing_on_connect(self, client, userdata, flags, rc):
        """ my own connect handling."""
        infostr = "My CONNACK received with code: {0}; client_obj: {1}.".format(rc, client)
        self.cfg_logging().info(infostr)
        print(infostr)
        if rc == 0:
            self._wait4connection = False
            self._client_handle(client)
            self.subscribe_settopics('hometop/ht/#')

    def processing_payload(self, userdata, topic, payload):
        """ """
        if 'set/' in topic:
            return
        else:
            if 'hc1_Tdesired' in topic:
#                print("topic:{0}: {1}".format(topic, payload))
                if ',' in payload:
                    #stripdown and get value only
                    stripped = payload.split(',')
                    value = stripped[0]
                else:
                    value = payload
                self.data.f_Soll_HK(value)
            if 'hc1_Tmeasured' in topic:
                self.data.f_Steuer(payload)
            if 'sol_Tcollector' in topic:
#                print("topic:{0}: {1}".format(topic, payload))
                self.data.f_kollektor(payload)
            if 'sol_Tcylinder_bottom' in topic:
                self.data.f_speicherunten(payload)
#                print("topic:{0} :{1}".format(topic, payload))
            if 'sol_pump' in topic:
                self.data.b_pumpe(payload)
#                print("topic:{0}: {1}".format(topic, payload))
            if 'ch_burner_operation' in topic:
                self.data.b_brennerflamme(payload)
            if 'hc1_Tniveau' in topic:
#                print("topic:{0}: {1}".format(topic, payload))
                self.data.i_temperatur_niveau(payload)
            if 'hc1_operationstatus' in topic:
                self.data.i_betriebsstatus(payload)
#                print("topic:{0}: {1}".format(topic, payload))

class csolarcontrol():
    global threadlock
    global g_TSolarSpeicherMaxValue
    global g_TWatermarkLow
    global g_TWatermarkHigh
    global g_THysterese
    global g_TSollDefault

    OP_STATUS_AUTO = "Auto"

    OP_NIVEAU_HEIZEN = "Heizen"

    OperationNiveau = {
        0:"Off",
        1:"Frost",
        2:"Sparen",
        3:OP_NIVEAU_HEIZEN,
        "Off":0,
        "Frost":1,
        "Sparen":2,
        OP_NIVEAU_HEIZEN:3
    }

    OperationStatus = {
        0:"Off",
        1:"Manuell",
        2:"Auto",
        3:"Urlaub",
        4:"E-Trocknung_4",
        5:"E-Trocknung_5",
        "Off":0,
        "Manuell":1,
        "Auto":2,
        "Urlaub":3,
        "E-Trocknung_4":4,
        "E-Trocknung_5":5,
    }

    def __init__(self, mqttclient_handle, logging, data):
        self.__threadrun = True
        self.__mqttclient = mqttclient_handle
        self.__logging = logging
        self.__data = data
       
    def __del__(self):
        self.stop()
        
    def stop(self):
#        self._mqtt_client.stop()
        self.__threadrun=False

    def __IsWatermarkAboveLow(self):
        if self.__data.f_speicherunten() > g_TWatermarkLow:
            return True
        else:
            return False
        
    def __IsWatermarkAboveHigh(self):
        if self.__data.f_speicherunten() > g_TWatermarkHigh:
            return True
        else:
            return False
        
    def __IsSolarSpeicherVoll(self):
        if self.__data.f_speicherunten() >= g_TSolarSpeicherMaxValue:
            return True
        else:
            return False

    def __IsSolarPowerAvailable(self):
        if self.__data.f_kollektor() > (self.__data.f_speicherunten() + g_THysterese):
            return True
        else:
            return False

    def __IsCollektorTempAboveWaterMark(self):
        if self.__data.f_kollektor() > g_TWatermarkLow:
            return True
        else:
            return False

    def __IsHeaterRunning(self):
        if self.__data.b_brennerflamme() == 1:
            return True
        else:
            return False

    def __IsInterceptionRequired(self):
        fktrtn=False
        if not self.__IsHeaterRunning():
            if self.__IsSolarSpeicherVoll() or self.__data.f_kollektor() > (g_TSolarSpeicherMaxValue + g_THysterese):
                fktrtn=True
            elif self.__IsSolarPowerAvailable() and self.__IsWatermarkAboveLow():
                fktrtn=True
        return fktrtn
    
    def __CurrentTSollHK1(self):
        tsollrtn=0
        if self.__data.f_Soll_HK() > 5.0 and self.__data.f_Soll_HK() < 31.0:
            tsollrtn = self.__data.f_Soll_HK()
        return tsollrtn

    def __calculateregelvalue(self, tcollector):
        rtnvalue=2
        maxvalue =max(tcollector)
        minvalue =min(tcollector)
        diffvalue=maxvalue-minvalue
        # check limits
        if maxvalue>0 and minvalue>0 and diffvalue>4:
            #check for rising edge
            if ((tcollector[0]+1.0) < tcollector[3]):
                rtnvalue=int(diffvalue / 4)*2
        return rtnvalue

    def __ControllingRequired(self):
        """check month for hot-range"""
        rtnvalue = False
        month = int(time.localtime().tm_mon)
        myhour = int(time.localtime().tm_hour)
#        print("Month: {0}; Time: {1}".format(month, myhour))
        if month > 3 and month < 11:
            if myhour > 9 and myhour < 19:
                rtnvalue = True
        return bool(rtnvalue)

    def __IsSetupHeizenRequired(self):
        """check the current NIVEAU and Status for requirement."""
        rtnvalue = False
        if "Cxyz" in g_ControllerType:
            #do nothing
            pass
        else:
#            print("betriebstatus:{0}; temp_niveau:{1}".format(self.__data.i_betriebsstatus(), self.__data.i_temperatur_niveau()))
            if self.__ControllingRequired() and self.__IsCollektorTempAboveWaterMark():

                if self.__data.i_betriebsstatus() == csolarcontrol.OperationStatus[csolarcontrol.OP_STATUS_AUTO]:
                    rtnvalue = True
                    debugstr = "setup Manuell required"
                    self.__logging.info(debugstr)
#                    print(debugstr)

                if self.__data.i_temperatur_niveau() != csolarcontrol.OperationNiveau[csolarcontrol.OP_NIVEAU_HEIZEN]:
                    rtnvalue = True
                    debugstr = "setup Heizen required"
                    self.__logging.info(debugstr)
#                    print(debugstr)
            else:
                rtnvalue = False
        return bool(rtnvalue)

    def run_solarcontrol(self):
        f_TSoll   =g_TSollDefault
        f_TSollOld=20.0
        SetupCounter=0
        SendRequest=False
        TCollectorArray=[1,2,3,4]
        TCollectorIndex=0
        addonvalue=2
        first_start = True

        self.__logging.info("csolarcontrol();start")

        try:
            while self.__threadrun:
                if self.__ControllingRequired():
                    if first_start:
                        first_start = False
                        time.sleep(60)
                    time.sleep(30)

                    # check for Heizen required
                    self.__IsSetupHeizenRequired()

                    #save 4 Tcollector-values
                    TCollectorArray[TCollectorIndex] = self.__data.f_kollektor()
                    TCollectorIndex+=1
                    if TCollectorIndex>3:
                        #calculate steigung and get addon-value
                        addonvalue = self.__calculateregelvalue(TCollectorArray)
                        TCollectorIndex=0
                
                    if self.__IsInterceptionRequired():
                        if self.__IsWatermarkAboveHigh():
                            f_TSoll=30.0
                        elif self.__data.f_speicherunten() > g_TWatermarkMed:
#zs#test#
                            f_TSoll=23.0+addonvalue
                            if f_TSoll > 30.0:
                                f_TSoll=30.0
                            
#                        f_TSoll=29.0
                        else:
                            f_TSoll=24.0
                    else:
                        f_TSoll=g_TSollDefault

                    str_csvlog=""";TCollect:{0};TSpeicher:{1};Solpump:{2};TRaumSoll_cmd:{3};TRaumSollHK1_ist:{4};TRaumIst:{5};""".format(self.__data.f_kollektor(),
                                                                    self.__data.f_speicherunten(),
                                                                    self.__data.b_pumpe(),
                                                                    f_TSoll,
                                                                    self.__CurrentTSollHK1(),
                                                                    self.__data.f_Steuer())

                    if (f_TSollOld != f_TSoll):
                        f_TSollOld  = f_TSoll
                        SetupCounter= 0
                        SendRequest = True

                    if SendRequest:
                        if SetupCounter < 10:
                            SetupCounter += 1
                        # ueberpruefung ob geforderter soll-wert ungleich aktuellem sollwert ist,
                        # falls ja dann neuen Sollwert zum Regler senden
                            if f_TSoll != self.__CurrentTSollHK1():
                            #send setup
                                str_csvlog = str_csvlog + "SetupCmd to heater;"
                                try:
                                    if "Cxyz" in g_ControllerType:
                                        #send temp- temperatur to broker
                                        topic = 'set/hometop/ht/hc1_Tdesired'
                                        payload = "{0},temporary".format(f_TSoll)
                                        self.__mqttclient.publish_data(topic, payload)
                                        time.sleep(8)
                                    else:
                                        #send temperatur-niveau: heizen to broker
                                        topic = 'set/hometop/ht/hc1_Tniveau'
                                        payload = "heizen"
                                        self.__mqttclient.publish_data(topic, payload)
                                        time.sleep(8)
                                        topic = 'set/hometop/ht/hc1_Tdesired'
                                        payload = "{0},heizen".format(f_TSoll)
                                        self.__mqttclient.publish_data(topic, payload)
                                        time.sleep(8)
                                except:
                                    str_error=("csolarcontrol();could not set temperatur-level: {0}".format(error))
                                    self.__logging.critical(str_error)
#                                    print(str_error)
                            else:
                            #kein bedarf
                                SendRequest  = False
                                SetupCounter = 0
                        else:
                            SendRequest  = False
                            SetupCounter = 0
                    self.__logging.info(str_csvlog)
                else:
                    time.sleep(60)
                    if "Cxyz" in g_ControllerType:
                        topic = 'set/hometop/ht/hc1_Tdesired'
                        payload = "{0},temporary".format(g_TSollDefault)
                        self.__mqttclient.publish_data(topic, payload)
                        time.sleep(8)
                    else:
                        if self.__data.i_betriebsstatus() != csolarcontrol.OperationStatus[csolarcontrol.OP_STATUS_AUTO]:
                            debugstr = "setup Auto required"
                            self.__logging.info(debugstr)
                            topic = 'set/hometop/ht/hc1_Tniveau'
                            payload = "auto"
                            try:
                                self.__mqttclient.publish_data(topic, payload)
                            except:
                                debugstr = "publish topic {0}: {1}".format(topic, payload)
                                str_error="csolarcontrol();error on: " + debugstr
                                self.__logging.critical(str_error)
                            else:
                                time.sleep(8)
        except:
            str_error="csolarcontrol();error in fkt:'run()'"
            self.__logging.critical(str_error)
#            print(str_error)
            self.stop()
            raise
# end class csolarcontrol()

configfile="./etc/config/mqtt_client_cfg_solarcontrol.xml"
mqtt_client=my_cmqtt_client(configfile)
mqtt_client.run()
